#include <algorithm>
#include <cmath>
#include "SPH.h"
#include <ppl.h>

using namespace Concurrency;
using namespace Eigen;

/* Constructor */
SPH::SPH(int sceneNumber)
{   
  stepCount_ = 0u;
  sceneNumber_ = sceneNumber;
  InitializeParameters();  
  ComputeKernels();
  CreateVolumes();    
  CreateGrid(1.0f);
  InitializeRendering();  
}

/* Destructor */
SPH::~SPH()
{
  delete [] allParticles_;
  delete [] neighbors_;
  delete [] grid_;
  delete [] gridCount_;
  glDeleteLists(sphere_, 1);
  glDeleteLists(ball_, 1);
}

/* Initializes rendering stuffs */
void SPH::InitializeRendering()
{
  sphere_ = glGenLists(1);
  glNewList(sphere_, GL_COMPILE);
  glutSolidSphere(0.5, 12, 12);
  glEndList();

  ball_ = glGenLists(1);
  glNewList(ball_, GL_COMPILE);
  glutSolidSphere(ballRadius_, 12, 12);
  glEndList();
}

/* Draws the particles on screen */
void SPH::DrawParticles()
{
  for (unsigned int i = 0u; i < particleCount_; ++i)
  {
    Particle& p = allParticles_[i];

    glPushMatrix();
    glTranslatef(p.position[0], p.position[1], p.position[2]);
    glCallList(sphere_);
    glPopMatrix();
  }  
}

void SPH::DrawBall()
{
  glPushMatrix();
  glTranslatef(ballCenter_[0], ballCenter_[1], ballCenter_[2]);
  glCallList(ball_);
  glPopMatrix();
}

/* Initializes the constants */
void SPH::InitializeParameters()
{
  /* default parameters
     sources of the values are provided whenever applicable */
  particleCount_ = 6400u; // increasing/decreasing this requires decreasing/increasing the kernelRadius_ as well
  restDensity_ = 1000.0f; // from en.wikipedia.org/wiki/Density#Density_of_water_.28at_1_atm.29
  pressureStiffness_ = 1.2f; // increasing this value to make the water less compressible
  viscosityCoefficient_ = 0.5f; // increasing this value to make the system more stable
  gravitationalAcceleration_ = -9.81f; // well-known value
  kernelRadius_ = 0.012f; // this is dependent on the particleCount (see above)
  kernelRadiusSquare_ = kernelRadius_ * kernelRadius_; // this is pre-computed since we use it very often
  timeStep_ = 0.004f;
  simulationScale_ = 0.004f; // from developer.amd.com/zones/OpenCLZone/Events/assets/SmoothedParticleHydrodynamics.pdf
  wallStiffness_ = 10000.0f;
  wallDamp_ = 250.0f;
  accelerationLimit_ = 200.0f;

  switch (sceneNumber_)
  {
  case 1: // dam break
    break;
  case 2: // water drop
    break;
  case 3: // sink
    wallStiffness_ = 20000.0f;
    break;
  case 4: // wave
    break;
  case 5: // ball
    ballCenter_ = Vector4f(0.0f, 400.0f, 0.0f, 0.0f);
    ballRadius_ = 10.0f;
    break;
  case 6: // U
    restDensity_ = 512.0f;
    kernelRadius_ = 0.01f;
    //wallStiffness_ = 10000.0f;    
    viscosityCoefficient_ = 0.6f;
    accelerationLimit_ = 32.0f;
    break;
  }
}

/* Projects the particles into the grid */
void SPH::InsertParticlesToGrid()
{    
  memset(gridCount_, 0, cellCount_ * sizeof(int));

  for (unsigned int i = 0u; i < particleCount_; ++i)
  {
    Particle& p = allParticles_[i]; 

    Vector4f coordinate = (p.position - minGridVolume_) / realCellSize_;    
    p.zIndex = Ultilities::EncodeMorton((unsigned int)coordinate[0], (unsigned int)coordinate[1], (unsigned int)coordinate[2]);
    
    if (p.zIndex < 0u || p.zIndex >= cellCount_)
      continue;

    if (gridCount_[p.zIndex] >= grid_[p.zIndex].size())
      grid_[p.zIndex].push_back(&p);
    else
      grid_[p.zIndex][gridCount_[p.zIndex]] = &p;

    gridCount_[p.zIndex]++;
  }
}

/* Initializes the grid */
void SPH::CreateGrid(float border)
{  
  minGridVolume_ = minTotalVolume_ - Vector4f(border, border, border, 0.0f);
  maxGridVolume_ = maxTotalVolume_ + Vector4f(border, border, border, 0.0f);  

  realCellSize_ = cellSize_ / simulationScale_;  
      
  gridResolution_ = (maxGridVolume_ - minGridVolume_) / realCellSize_;
  gridResolution_[0] = ceil(gridResolution_[0]);
  gridResolution_[1] = ceil(gridResolution_[1]);
  gridResolution_[2] = ceil(gridResolution_[2]);  
  
  cellCount_ = Ultilities::EncodeMorton((unsigned int) gridResolution_[0], (unsigned int) gridResolution_[1], (unsigned int) gridResolution_[2]);

  grid_ = new ParticleList[cellCount_];
  gridCount_ = new unsigned int[cellCount_];
}

/* Creates the intial volume */
void SPH::CreateVolumes()
{
  float volume;
  float gap;
  unsigned int particleCount1;
  Vector4f minInitialVolume;
  Vector4f maxInitialVolume;
  Vector4f minInitialVolume1;
  Vector4f maxInitialVolume1;
  Vector4f minInitialVolume2;
  Vector4f maxInitialVolume2;  
  minTotalVolume_ = Vector4f(-32.0f, 0.0f, -16.0f, 0.0f);  
  maxTotalVolume_ = Vector4f(32.0f, 64.0f, 16.0f, 0.0f);

  switch (sceneNumber_)
  {
  case 1: // dam break    
    minInitialVolume = Vector4f(0.0f, 0.0f, -15.0f, 0.0f);  
    maxInitialVolume = Vector4f(31.0f, 32.0f, 15.0f, 0.0f);  
    volume = (maxInitialVolume[0] - minInitialVolume[0]) * simulationScale_ *
             (maxInitialVolume[1] - minInitialVolume[1]) * simulationScale_ *
             (maxInitialVolume[2] - minInitialVolume[2]) * simulationScale_;  
    particleMass_ = restDensity_ * volume / (float)particleCount_;
    gap = pow(particleMass_ / restDensity_, 1.0f / 3.0f) / simulationScale_;
    particleCount_ = CountParticles(minInitialVolume, maxInitialVolume, gap);
    allParticles_ = new Particle[particleCount_];
    neighbors_ = new ParticleList[particleCount_];
    InsertParticlesToVolume(0u, minInitialVolume, maxInitialVolume, gap);          
    break;
  case 2: // water drop    
    minInitialVolume1 = Vector4f(-31.0f, 0.0f, -15.0f, 0.0f);  
    maxInitialVolume1 = Vector4f(31.0f, 5.0f, 15.0f, 0.0f);  
    volume = (maxInitialVolume1[0] - minInitialVolume1[0]) * simulationScale_ *
             (maxInitialVolume1[1] - minInitialVolume1[1]) * simulationScale_ *
             (maxInitialVolume1[2] - minInitialVolume1[2]) * simulationScale_;  
    minInitialVolume2 = Vector4f(-10.0f, 20.0f, -4.0f, 0.0f);  
    maxInitialVolume2 = Vector4f(10.0f, 59.0f, 4.0f, 0.0f);
    volume += (maxInitialVolume2[0] - minInitialVolume2[0]) * simulationScale_ *
              (maxInitialVolume2[1] - minInitialVolume2[1]) * simulationScale_ *
              (maxInitialVolume2[2] - minInitialVolume2[2]) * simulationScale_;  
    particleMass_ = restDensity_ * volume / (float)particleCount_;
    gap = pow(particleMass_ / restDensity_, 1.0f / 3.0f) / simulationScale_;    
    particleCount_ = CountParticles(minInitialVolume1, maxInitialVolume1, gap);
    particleCount1 = particleCount_;
    particleCount_ += CountParticles(minInitialVolume2, maxInitialVolume2, gap);
    allParticles_ = new Particle[particleCount_];
    neighbors_ = new ParticleList[particleCount_];        
    InsertParticlesToVolume(0u, minInitialVolume1, maxInitialVolume1, gap);
    InsertParticlesToVolume(particleCount1, minInitialVolume2, maxInitialVolume2, gap);
    break;
  case 3: // 2 levels
    minInitialVolume = Vector4f(-31.0f, 34.0f, -15.0f, 0.0f);  
    maxInitialVolume = Vector4f(31.0f, 59.0f, 15.0f, 0.0f);  
    volume = (maxInitialVolume[0] - minInitialVolume[0]) * simulationScale_ *
             (maxInitialVolume[1] - minInitialVolume[1]) * simulationScale_ *
             (maxInitialVolume[2] - minInitialVolume[2]) * simulationScale_;  
    particleMass_ = restDensity_ * volume / (float)particleCount_;
    gap = pow(particleMass_ / restDensity_, 1.0f / 3.0f) / simulationScale_;
    particleCount_ = CountParticles(minInitialVolume, maxInitialVolume, gap);
    allParticles_ = new Particle[particleCount_];
    neighbors_ = new ParticleList[particleCount_];
    InsertParticlesToVolume(0u, minInitialVolume, maxInitialVolume, gap);          
    break;
  case 4: // wave
    minInitialVolume = Vector4f(-31.0f, 0.0f, -15.0f, 0.0f);  
    maxInitialVolume = Vector4f(0.0f, 32.0f, 15.0f, 0.0f);  
    volume = (maxInitialVolume[0] - minInitialVolume[0]) * simulationScale_ *
             (maxInitialVolume[1] - minInitialVolume[1]) * simulationScale_ *
             (maxInitialVolume[2] - minInitialVolume[2]) * simulationScale_;  
    particleMass_ = restDensity_ * volume / (float)particleCount_;
    gap = pow(particleMass_ / restDensity_, 1.0f / 3.0f) / simulationScale_;
    particleCount_ = CountParticles(minInitialVolume, maxInitialVolume, gap);
    allParticles_ = new Particle[particleCount_];
    neighbors_ = new ParticleList[particleCount_];
    InsertParticlesToVolume(0u, minInitialVolume, maxInitialVolume, gap);          
    break;
  case 5: // ball
    minInitialVolume = Vector4f(-31.0f, 0.0f, -15.0f, 0.0f);  
    maxInitialVolume = Vector4f(31.0f, 10.0f, 15.0f, 0.0f);  
    volume = (maxInitialVolume[0] - minInitialVolume[0]) * simulationScale_ *
             (maxInitialVolume[1] - minInitialVolume[1]) * simulationScale_ *
             (maxInitialVolume[2] - minInitialVolume[2]) * simulationScale_;              
    particleMass_ = restDensity_ * volume / (float)particleCount_;
    gap = pow(particleMass_ / restDensity_, 1.0f / 3.0f) / simulationScale_;    
    particleCount_ = CountParticles(minInitialVolume, maxInitialVolume, gap);        
    allParticles_ = new Particle[particleCount_];
    neighbors_ = new ParticleList[particleCount_];
    InsertParticlesToVolume(0u, minInitialVolume1, maxInitialVolume1, gap);    
    break;
  case 6: // U
    minInitialVolume = Vector4f(23.0f, 1.0f, -15.0f, 0.0f);  
    maxInitialVolume = Vector4f(31.0f, 59.0f, 15.0f, 0.0f);  
    volume = (maxInitialVolume[0] - minInitialVolume[0]) * simulationScale_ *
             (maxInitialVolume[1] - minInitialVolume[1]) * simulationScale_ *
             (maxInitialVolume[2] - minInitialVolume[2]) * simulationScale_;              
    particleMass_ = restDensity_ * volume / (float)particleCount_;
    gap = pow(particleMass_ / restDensity_, 1.0f / 3.0f) / simulationScale_;    
    particleCount_ = CountParticles(minInitialVolume, maxInitialVolume, gap);        
    allParticles_ = new Particle[particleCount_];
    neighbors_ = new ParticleList[particleCount_];
    InsertParticlesToVolume(0u, minInitialVolume1, maxInitialVolume1, gap);    
    break;
  }
    
  particleRadius_ = pow(3.0f * particleMass_ / (4.0f * PI * restDensity_), 1.0f / 3.0f);
  kernelRadius_ *= 1.2f;
  kernelRadiusSquare_ = kernelRadius_ * kernelRadius_;
  cellSize_ = kernelRadius_;  
}

/* Counts the number of particles in the volume specified by min and max */
unsigned int SPH::CountParticles(const Vector4f& min, const Vector4f& max, float gap)
{
  unsigned int count = 0u;
  for (float x = min[0]; x <= max[0]; x += gap)   
    for (float y = min[1]; y <= max[1]; y += gap)    
      for (float z = min[2]; z <= max[2]; z += gap)              
        ++count;
  return count;
}

/* Creates the particles at their initial positions */
void SPH::InsertParticlesToVolume(unsigned int begin, const Vector4f& minVolume, const Vector4f& maxVolume, float gap)
{        
  unsigned int index = begin;
  for (float x = minVolume[0]; x <= maxVolume[0]; x += gap) 
  {
    for (float y = minVolume[1]; y <= maxVolume[1]; y += gap)
    {
      for (float z = minVolume[2]; z <= maxVolume[2]; z += gap)
      {
        Particle& p = allParticles_[index];
        p.id = index;
        p.density = p.pressure = 0.0f;
        p.acceleration.setZero();
        p.velocity.setZero();
        p.prevVelocity.setZero();
        p.position = Vector4f(x, y, z, 0.0f);
        p.neighborCount = 0u;

        ++index;
      }
    }
  }  
}

/* Precomputes the kernels using [Muller 2003] */
void SPH::ComputeKernels()
{
  WPoly6 = 315.0f / (64.0f * PI * pow(kernelRadius_, 9.0f));
  GWSpiky = -45.0f / (PI * pow(kernelRadius_, 6.0f));
  LWViscosity = 45.0f / (PI * pow(kernelRadius_, 6.0f));
}

/* Calculates pressure on particles using [Muller 2003] 
   and populates the neighbor table at the same time */
void SPH::ComputePressure()
{
  parallel_for(0u, particleCount_, 1u, [&](unsigned int i)
  {        
    Particle& p = allParticles_[i];
    p.neighborCount = 0u;
    Vector4i coord = ((p.position - minGridVolume_) / realCellSize_).cast<int>();            
    float sum = 0.0f;
    for (int dx = -1; dx <= 1; ++dx)
    {
      for (int dy = -1; dy <= 1; ++dy)
      {
        for (int dz = -1; dz <= 1; ++dz)
        {
          Vector4i ncoord = coord + Vector4i(dx, dy, dz, 0);            
          if (ncoord[0] < 0 || ncoord[0] >= (int)gridResolution_[0] || 
              ncoord[1] < 0 || ncoord[1] >= (int)gridResolution_[1] ||
              ncoord[2] < 0 || ncoord[2] >= (int)gridResolution_[2])
            continue;
          unsigned int nIdx = (unsigned int)Ultilities::EncodeMorton((unsigned int)ncoord[0], (unsigned int)ncoord[1], (unsigned int)ncoord[2]);         
          ParticleList& neighborList = grid_[nIdx];               
          for (unsigned int j = 0; j < gridCount_[nIdx]; ++j)
          {
            Particle& q = *(neighborList[j]);
            if (&p == &q) continue;
            Vector4f diff = p.position * simulationScale_ - q.position * simulationScale_;              
            float distanceSquare = diff.squaredNorm();
            if (distanceSquare <= kernelRadiusSquare_)
            {
              float hr = kernelRadiusSquare_ - distanceSquare;
              sum += hr * hr * hr;              
              if (p.neighborCount >= neighbors_[p.id].size())  
                neighbors_[p.id].push_back(&q);
              else
                neighbors_[p.id][p.neighborCount] = &q;
              p.neighborCount++;
            }
          }          
        }
      }
    }
    p.density = sum * WPoly6 * particleMass_;    
    p.pressure = pressureStiffness_ * (p.density - restDensity_);      
  });
}

/* Uses a stiff spring force with damped motion to push the particles away from the walls
   The method is described at gafferongames.com/game-physics/spring-physics/ */
inline void SPH::ResponseToCollisions(const Vector4f& velocity, float diff, const Vector4f& wallNormal, Vector4f& acceleration)
{
  float dot = wallNormal.dot(velocity);  
  acceleration += wallNormal * (wallStiffness_ * diff - wallDamp_ * dot);
}

/* Detects collisions between particle p and the boundary planes */
void SPH::DetectCollisions(Particle& p)
{           
  static float epsilon = 0.00001f;
  // min X
  float diff = particleRadius_ - (p.position[0] - minTotalVolume_[0]) * simulationScale_;  
  if (diff > epsilon)      
    ResponseToCollisions(p.velocity, diff, Vector4f(1.0f, 0.0f, 0.0f, 0.0f), p.acceleration);  
  // max X
  diff = particleRadius_ - (maxTotalVolume_[0] - p.position[0]) * simulationScale_;  
  if (diff > epsilon)
    ResponseToCollisions(p.velocity, diff, Vector4f(-1.0f, 0.0f, 0.0f, 0.0f), p.acceleration);  
  // min Y
  diff = particleRadius_ - (p.position[1] - minTotalVolume_[1]) * simulationScale_;  
  if (diff > epsilon)
    ResponseToCollisions(p.velocity, diff, Vector4f(0.0f, 1.0f, 0.0f, 0.0f), p.acceleration);  
  // max Y
  diff = particleRadius_ - (maxTotalVolume_[1] - p.position[1]) * simulationScale_;
  if (diff > epsilon)
    ResponseToCollisions(p.velocity, diff, Vector4f(0.0f, -1.0f, 0.0f, 0.0f), p.acceleration);    
  // min Z
  diff = particleRadius_ - (p.position[2] - minTotalVolume_[2]) * simulationScale_;
  if (diff > epsilon)
    ResponseToCollisions(p.velocity, diff, Vector4f(0.0f, 0.0f, 1.0f, 0.0f), p.acceleration);      
  // max Z
  diff = particleRadius_ - (maxTotalVolume_[2] - p.position[2]) * simulationScale_;
  if (diff > epsilon)
    ResponseToCollisions(p.velocity, diff, Vector4f(0.0f, 0.0f, -1.0f, 0.0f), p.acceleration);          

  float dist;  
  switch (sceneNumber_)
  {
  case 1: // dam break
    break;
  case 2: // water drop
    break;
  case 3: // sink
    dist = p.position[0] * p.position[0] + p.position[2] * p.position[2];
    if (p.position[1] > 32.0f && dist >= 64.0f)
    {
      diff = particleRadius_ - (p.position[1] - 33.0f) * simulationScale_;  
      if (diff > epsilon)
        ResponseToCollisions(p.velocity, diff, Vector4f(0.0f, 1.0f, 0.0f, 0.0f), p.acceleration);   
    }
    break;
  case 4: // wave    
    diff = particleRadius_ - (wallX_ - p.position[0]) * simulationScale_;
    if (diff > epsilon)
      ResponseToCollisions(p.velocity, diff, Vector4f(-1.0f, 0.0f, 0.0f, 0.0f), p.acceleration);          
    break;
  case 5: // ball
    diff = particleRadius_ - ((p.position - ballCenter_).norm() - ballRadius_) * simulationScale_;
    if (diff > epsilon)                
      ResponseToCollisions(p.velocity, diff, (p.position - ballCenter_).normalized(), p.acceleration);              
    break;
  case 6: // U    
    if (p.position[1] >= 6.0f && p.position[0] >= 22.0f)
    {
      diff = particleRadius_ - (p.position[0] - 22.0f) * simulationScale_;
      if (diff > epsilon)                
        ResponseToCollisions(p.velocity, diff, Vector4f(1.0f, 0.0f, 0.0f, 0.0f), p.acceleration);                    
    }
    else if (p.position[1] >= 6.0f && p.position[0] <= -22.0f)
    {
      diff = particleRadius_ - (-22.0f - p.position[0]) * simulationScale_;
      if (diff > epsilon)                
        ResponseToCollisions(p.velocity, diff, Vector4f(-1.0f, 0.0f, 0.0f, 0.0f), p.acceleration);              
    }
    else if (p.position[1] <= 6.0f && (p.position[0] <= 22.0f && p.position[0] >= -22.0f))
    {
      diff = particleRadius_ - (6.0f - p.position[1]) * simulationScale_;
      if (diff > epsilon)                
        ResponseToCollisions(p.velocity, diff, Vector4f(0.0f, -1.0f, 0.0f, 0.0f), p.acceleration);              
    }
    break;
  }  

  // leap-frog integration  
  Vector4f nextVelocity = p.acceleration * timeStep_ + p.prevVelocity;  
  p.velocity = (p.prevVelocity + nextVelocity) * 0.5f;  
  p.prevVelocity = nextVelocity;  
  p.position += nextVelocity * (timeStep_ / simulationScale_);  
}

/* Sorts the particle array using insertion sort */
void SPH::SortParticles()
{
  for (unsigned int i = 1u; i < particleCount_; ++i)
  {
    Particle p = allParticles_[i];
    unsigned int j = i - 1u;
    bool done = false;
    while (!done)
    {
      if (allParticles_[j].zIndex > p.zIndex)
      {
        allParticles_[j + 1u] = allParticles_[j];
        if (j == 0u)
          done = true;        
        j = j - 1u;        
      }
      else
      {
        done = true;
      }
    }
    allParticles_[j + 1u] = p;
  }
}

/* Simulation step */
void SPH::TimeStep()
{    
  DrawParticles();

  InsertParticlesToGrid();  

  ComputePressure();

  if (sceneNumber_ == 4) // wave
  {
    wallX_ = 15.0f + sin((float)stepCount_ * 3.14159265f / 180.0f) * 10.0f;
  }
  else if (sceneNumber_ == 5) // ball
  {
    if (ballCenter_[1] - ballRadius_ > minTotalVolume_[1])    
      ballCenter_[1] -= 4.0f;    
    else
      ballCenter_[0] = 0.0f + sin((float)stepCount_ * 3.14159265f / 180.0f) * 20.0f;;
    DrawBall();
  }

  // compute accelerations due to viscosity, pressure, and gravity
  parallel_for(0u, particleCount_, 1u, [&](unsigned int i)  
  {
    Particle& p = allParticles_[i];
    p.acceleration.setZero();

    ParticleList& neighbors = neighbors_[p.id];      
    for (unsigned int j = 0; j < p.neighborCount; ++j)
    {
      Particle& q = *(neighbors[j]);
      Vector4f diff = p.position * simulationScale_ - q.position * simulationScale_;        
      float distance = diff.norm();      
      float hr = kernelRadius_ - distance;      
      float viscosityTerm = LWViscosity * viscosityCoefficient_;
      float pressureTerm = -0.5f * hr * GWSpiky * (p.pressure + q.pressure) / distance;      
      float commonTerm = particleMass_ * hr / (p.density * q.density);
      Vector4f viscosityAcceleration = (q.velocity - p.velocity) * (viscosityTerm * commonTerm);        
      Vector4f pressureAcceleration = diff * (pressureTerm * commonTerm);        
      p.acceleration += viscosityAcceleration + pressureAcceleration;        
    }   

    // damp the acceleration if it gets too large
    float accelerationSquaredNorm = p.acceleration.squaredNorm();
    if (accelerationSquaredNorm > accelerationLimit_ * accelerationLimit_)  
      p.acceleration *= accelerationLimit_ / sqrt(accelerationSquaredNorm);

    // add the gravitational acceleration
    // must be added after the acceleration is damped to avoid dealing with infinity values
    p.acceleration[1] += gravitationalAcceleration_;
  });  
    
  // detect boundary collision and advances the particles
  parallel_for(0u, particleCount_, 1u, [&](unsigned int i)
  {
    Particle& p = allParticles_[i];
    DetectCollisions(p); 
  });     

  // sort the particles every 100th simulation step
  static unsigned int stepCount = 0u;
  if (stepCount == 100u)
  {
    SortParticles();
    stepCount = 0u;
  }
  stepCount++;

  // this counter is only used in some scenes
  if (stepCount_ == 360u)
    stepCount_ = 0u;
  stepCount_ += 2u;
}